// SerialIO.cpp

// Fran Jarnjak
// June 2007
// fjarnjak_at_gmail_dot_com

// Implementation file of SerialIO class


#include "SerialIO.h"
#include <iostream>
#include <stdexcept>
#include <time.h>
#include <fcntl.h>
#include <errno.h>

using namespace std;

SerialIO::SerialIO(const char* pth, int br, int db, int p) : path(pth), baud(br), dataBits(db), parity(p), valid(true)
{    
    try
    {
       openSerial();
    }catch(...)
    {
       cout << "Error opening serial port: " << strerror(errno) << "(" << errno << ")" << endl; 
        valid = false;
    }
}


SerialIO::~SerialIO()
{
    if(fd>0)
    {
    
      closeSerial();
        
      tcflush(fd, TCIOFLUSH);
      close(fd);
    }
    pthread_mutex_destroy(&writeMutex);
}

void SerialIO::closeSerial()
{
    tcflush(fd, TCIOFLUSH);
    if (tcsetattr(fd, TCSANOW, &oldFlags) < 0)
    {
	throw std::runtime_error("Can't set old terminal flags before closing");
    }
    close(fd);
    valid = false;
}

void SerialIO::openSerial()
{
 
        fd = open(path.c_str(), O_RDWR | O_NOCTTY | O_NONBLOCK);
    
	if (fd < 0)
	{
        std::string err = "Can't open serial port: ";
        err += path;
        throw std::runtime_error(err.c_str());
	}
		
	if(tcgetattr(fd, &termFlags))
	{
        throw std::runtime_error("Can't get terminal flags");
	}
        oldFlags = termFlags; 
    
        cout << "  Current in baud rate: " << cfgetispeed(&termFlags) << " bps" << endl;
	cout << "  Current out baud rate: " << cfgetospeed(&termFlags) << " bps" << endl;
	cout << "  Current c_cflag= " << termFlags.c_cflag << endl;
	cout << "  Current c_lflag= " << termFlags.c_lflag << endl;
	cout << "  Current c_iflag= " << termFlags.c_iflag << endl;
	cout << "  Current c_oflag= " << termFlags.c_oflag << endl;
        	
	termFlags.c_cflag = 0;
	termFlags.c_cflag |= (CS8 | CLOCAL | CREAD); 
	termFlags.c_lflag = 0; 
	termFlags.c_iflag = IGNBRK; 
	termFlags.c_oflag = 0; 
	
	termFlags.c_cc[VMIN] = 0;         /* enable read TimeOut */
	termFlags.c_cc[VTIME] = 40;       /* set 4 Second TimeOut */
	
	cfsetispeed(&termFlags, B38400);
	cfsetospeed(&termFlags, B38400);
	
	if (tcsetattr(fd, TCSANOW, &termFlags) < 0)
	{
		throw std::runtime_error("Can't set new terminal flags");
	}
	
	cout << ("After setting:\n");
	if(tcgetattr(fd, &termFlags))
	{
	    throw std::runtime_error("Can't get terminal flags after setting");
	}
    
	cout << "  Current in baud rate: " << cfgetispeed(&termFlags) << " bps" << endl;
	cout << "  Current out baud rate: " << cfgetospeed(&termFlags) << " bps" << endl;
	cout << "  Current c_cflag= " << termFlags.c_cflag << endl;
	cout << "  Current c_lflag= " << termFlags.c_lflag << endl;
	cout << "  Current c_iflag= " << termFlags.c_iflag << endl;
	cout << "  Current c_oflag= " << termFlags.c_oflag << endl;
	
       if(pthread_mutex_init(&writeMutex, NULL) != 0)
       {
          throw std::runtime_error("Can't init the writeMutex");
       }
}

// Write functions - returns result
void SerialIO::sendBytes(const char * data, int wsize, char * rslt, int * rsltSize)
{
    if(!valid) throw std::runtime_error("ModbusEngine not in a valid state");

    bool ok = true;
    int len = 0;
    char readBuff[256];
    
    // Mutex can be locked max 4 seconds. 
    pthread_mutex_lock(&writeMutex);
    
    if((len = write(fd, data, wsize)) <0)
	{
		cout << "Could not write all bytes " << strerror(errno) << "(" << errno << ")" << endl;
        ok = false;
 	}
    else
        cout << len << " bytes written." << endl;
    
	tcdrain(fd);
	
    len = 0;
    
    // Give chance to make reply - sleep for some time :)
    sleepTime(0.2); // sleep 0.2 seconds
    
	len = read(fd, readBuff, 256);
	
    if(len<=0)
    {
        cout << "ERROR reading. " << len << " << bytes read." << endl;
        ok - false;
    }
    else
        cout << len << " bytes read." << endl;
    
    pthread_mutex_unlock(&writeMutex);
    
    if(!ok)
        throw std::runtime_error("sending bytes failed");
    
    if(*rsltSize < len)
        throw std::runtime_error("More data read than passed buffer size");
    else
    {
        memcpy(rslt, readBuff, len);
        *rsltSize = len;
    }
    
}

void SerialIO::sleepTime(double fSec)
{
    timespec             tDelay;
    
    tDelay.tv_sec = (time_t)0;
    tDelay.tv_nsec = (long)(1000000L * (fSec*100));
    
    for(int i=0 ; i<10;i++)
        nanosleep(&tDelay, NULL);    
}
